use crate::ast::FunctionModifier::{Cons, Mut, MutRef, Ref, Static};
use crate::ast::*;

macro_rules! to_unparse_vec {
    ( $nodes:expr ) => {
        $nodes
            .iter()
            .map(|node| node as &dyn Unparse)
            .collect::<Vec<&dyn Unparse>>()
    };
}

macro_rules! unparse_contained_declarations {
    ( $declarations:expr, $buf:expr ) => {
        if !$declarations.is_empty() {
            write!($buf, "{{\n")?;
            unparse_list($buf, "\n\n", &to_unparse_vec!($declarations));
            write!($buf, "\n}}")?;
        }
    };
}

macro_rules! unparse_ok {
    () => {
        Ok(())
    };
}

pub trait Unparse {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result;
}

pub trait ListUnparse: Unparse {
    fn prefix() -> &'static str {
        ""
    }

    fn separator() -> &'static str;

    fn suffix() -> &'static str {
        ""
    }

    fn inner(&self) -> Vec<&dyn Unparse>;
}

impl<T: ListUnparse> Unparse for T {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "{}", Self::prefix())?;
        unparse_list(buf, Self::separator(), &self.inner())?;
        write!(buf, "{}", Self::suffix())?;
        unparse_ok!()
    }
}

fn unparse_list(
    buf: &mut dyn std::fmt::Write,
    sep: &str,
    items: &[&dyn Unparse],
) -> std::fmt::Result {
    for (i, item) in items.iter().enumerate() {
        item.unparse(buf)?;
        if i < items.len() - 1 {
            write!(buf, "{}", sep)?;
        }
    }
    Ok(())
}

fn unparse_comma_list(buf: &mut dyn std::fmt::Write, items: &[&dyn Unparse]) -> std::fmt::Result {
    unparse_list(buf, ", ", items)
}

/* Implementations */

/* Operators */

impl Unparse for Operator {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use Operator::*;
        match self {
            Binary(op) => op.unparse(buf),
            PrefixUnary(op) => op.unparse(buf),
            SuffixUnary(op) => op.unparse(buf),
        }
    }
}

impl Unparse for BinaryOperator {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use BinaryOperator::*;
        match self {
            Or => write!(buf, "||"),
            And => write!(buf, "&&"),
            EqualTo => write!(buf, "=="),
            NotEqualTo => write!(buf, "!="),
            Greater => write!(buf, ">"),
            Less => write!(buf, "<"),
            GreaterEqual => write!(buf, ">="),
            LessEqual => write!(buf, "<="),
            Add => write!(buf, "+"),
            Subtract => write!(buf, "-"),
            Multiply => write!(buf, "*"),
            Divide => write!(buf, "/"),
            Modulo => write!(buf, "%"),
            LeftShift => write!(buf, "<<"),
            RightShift => write!(buf, ">>"),
        }
    }
}

impl Unparse for PrefixUnaryOperator {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use PrefixUnaryOperator::*;
        match self {
            Spread => write!(buf, "..."),
            BorrowMut => write!(buf, "&mut"),
            Borrow => write!(buf, "&"),
            Mut => write!(buf, "mut"),
            Not => write!(buf, "!"),
            Negative => write!(buf, "-"),
        }
    }
}

impl Unparse for SuffixUnaryOperator {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use SuffixUnaryOperator::*;
        match self {
            PlusPlus => write!(buf, "++"),
            MinusMinus => write!(buf, "--"),
        }
    }
}

/* Names */

impl Unparse for Identifier {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "{}", self.name)
    }
}

impl ListUnparse for FullyQualifiedName {
    fn separator() -> &'static str {
        "::"
    }

    fn inner(&self) -> Vec<&dyn Unparse> {
        to_unparse_vec!(self.identifiers)
    }
}

/* TypeUse and components */

impl Unparse for TypeUse {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use TypeUse::*;
        match self {
            Void => write!(buf, "Void"),
            InterfaceOrClass(interface_or_class_type_use) => {
                interface_or_class_type_use.unparse(buf)
            }
            Tuple(tuple_type_use) => tuple_type_use.unparse(buf),
            Function(function_type_use) => function_type_use.unparse(buf),
        }
    }
}

impl Unparse for InterfaceOrClassTypeUse {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        for _ in 0..self.borrow_count {
            write!(buf, "&")?;
        }
        if self.is_mutable {
            write!(buf, "mut")?;
        }
        if self.borrow_count > 0 || self.is_mutable {
            write!(buf, " ")?;
        }
        self.fqn.unparse(buf)?;
        self.generics.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for TupleTypeUse {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        for _ in 0..self.borrow_count {
            write!(buf, "&")?;
        }
        if self.is_mutable {
            write!(buf, "mut")?;
        }
        if self.borrow_count > 0 || self.is_mutable {
            write!(buf, " ")?;
        }
        self.arguments.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for FunctionTypeUse {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        for _ in 0..self.borrow_count {
            write!(buf, "&")?;
        }
        if let Some(function_type_modifier) = &self.function_modifier {
            function_type_modifier.unparse(buf)?;
        }
        if self.borrow_count > 0 || self.function_modifier.is_some() {
            write!(buf, " ")?;
        }
        write!(buf, "fn ")?;
        if !self.generics.is_empty() {
            self.generics.unparse(buf)?;
            write!(buf, " ")?;
        }
        self.parameters.unparse(buf)?;
        write!(buf, " ")?;
        self.return_type.unparse(buf)?;
        unparse_ok!()
    }
}

/* Generic arguments */

impl ListUnparse for GenericArguments {
    fn prefix() -> &'static str {
        "<"
    }

    fn separator() -> &'static str {
        ", "
    }

    fn suffix() -> &'static str {
        ">"
    }

    fn inner(&self) -> Vec<&dyn Unparse> {
        to_unparse_vec!(self.0)
    }
}

/* Generic Parameters */

impl ListUnparse for GenericParameters {
    fn prefix() -> &'static str {
        "<"
    }

    fn separator() -> &'static str {
        ", "
    }

    fn suffix() -> &'static str {
        ">"
    }

    fn inner(&self) -> Vec<&dyn Unparse> {
        to_unparse_vec!(self.0)
    }
}

/* Tuple Arguments */

impl ListUnparse for TupleArguments {
    fn prefix() -> &'static str {
        "("
    }

    fn separator() -> &'static str {
        ", "
    }

    fn suffix() -> &'static str {
        ")"
    }

    fn inner(&self) -> Vec<&dyn Unparse> {
        to_unparse_vec!(self.0)
    }
}

/* Implements List */

impl ListUnparse for ImplementsList {
    fn prefix() -> &'static str {
        ": "
    }

    fn separator() -> &'static str {
        " + "
    }

    fn inner(&self) -> Vec<&dyn Unparse> {
        to_unparse_vec!(self.0)
    }
}

/* Function Type Modifier */

impl Unparse for FunctionTypeModifier {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use FunctionTypeModifier::*;
        match self {
            Cons => write!(buf, "cons"),
            MutRef => write!(buf, "mut ref"),
            Mut => write!(buf, "mut"),
            Ref => write!(buf, "ref"),
        }
    }
}

/* Parameters */

impl ListUnparse for Parameters {
    fn prefix() -> &'static str {
        "("
    }

    fn separator() -> &'static str {
        ", "
    }

    fn suffix() -> &'static str {
        ")"
    }

    fn inner(&self) -> Vec<&dyn Unparse> {
        to_unparse_vec!(self.0)
    }
}

impl Unparse for Parameter {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        self.identifier.unparse(buf)?;
        write!(buf, ": ")?;
        self.type_use.unparse(buf)?;
        unparse_ok!()
    }
}

/* Return Type */

impl Unparse for ReturnType {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "-> ")?;
        self.declared_type.unparse(buf)?;
        if !self.references.is_empty() {
            write!(buf, " ")?;
            self.references.unparse(buf)?;
        }
        unparse_ok!()
    }
}

impl ListUnparse for References {
    fn prefix() -> &'static str {
        "ref "
    }

    fn separator() -> &'static str {
        ", "
    }

    fn inner(&self) -> Vec<&dyn Unparse> {
        to_unparse_vec!(self.0)
    }
}

/* Top-level construct */

impl Unparse for CompilationUnit {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if let Some(namespace) = &self.namespace {
            write!(buf, "ns ")?;
            namespace.unparse(buf)?;
            write!(buf, "\n\n")?;
        }
        unparse_list(buf, "\n\n", &to_unparse_vec!(self.declarations))?;
        unparse_ok!()
    }
}

/* Declarations allowed in each level */

impl Unparse for ModuleLevelDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use ModuleLevelDeclaration::*;
        match self {
            Module(module_declaration) => module_declaration.unparse(buf),
            Interface(interface_declaration) => interface_declaration.unparse(buf),
            Class(class_declaration) => class_declaration.unparse(buf),
            Function(function_declaration) => function_declaration.unparse(buf),
            PlatformFunction(platform_function_declaration) => {
                platform_function_declaration.unparse(buf)
            }
        }
    }
}

impl Unparse for InterfaceLevelDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use InterfaceLevelDeclaration::*;
        match self {
            Module(module_declaration) => module_declaration.unparse(buf),
            Interface(interface_declaration) => interface_declaration.unparse(buf),
            Class(class_declaration) => class_declaration.unparse(buf),
            Function(interface_function_declaration) => interface_function_declaration.unparse(buf),
            OperatorFunction(interface_operator_function_declaration) => {
                interface_operator_function_declaration.unparse(buf)
            }
        }
    }
}

impl Unparse for ClassLevelDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use ClassLevelDeclaration::*;
        match self {
            Module(module_declaration) => module_declaration.unparse(buf),
            Interface(interface_declaration) => interface_declaration.unparse(buf),
            Class(class_declaration) => class_declaration.unparse(buf),
            Function(function_declaration) => function_declaration.unparse(buf),
            OperatorFunction(operator_function_declaration) => {
                operator_function_declaration.unparse(buf)
            }
            PlatformFunction(platform_function_declaration) => {
                platform_function_declaration.unparse(buf)
            }
            Property(property_declaration) => property_declaration.unparse(buf),
            Field(field_declaration) => field_declaration.unparse(buf),
        }
    }
}

/* Declarations */

impl Unparse for ModuleDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if self.is_public {
            write!(buf, "pub ")?;
        }
        write!(buf, "mod ")?;
        self.identifier.unparse(buf)?;
        unparse_contained_declarations!(self.declarations, buf);
        unparse_ok!()
    }
}

impl Unparse for InterfaceDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if self.is_public {
            write!(buf, "pub ")?;
        }
        write!(buf, "int ")?;
        self.identifier.unparse(buf)?;
        if !self.generics.is_empty() {
            self.generics.unparse(buf)?;
            write!(buf, " ")?;
        }
        if !self.implements.is_empty() {
            self.implements.unparse(buf)?;
            write!(buf, " ")?;
        }
        unparse_contained_declarations!(self.declarations, buf);
        unparse_ok!()
    }
}

impl Unparse for ClassDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if self.is_public {
            write!(buf, "pub ")?;
        }
        write!(buf, "class ")?;
        self.identifier.unparse(buf)?;
        if !self.generics.is_empty() {
            self.generics.unparse(buf)?;
        }
        if let Some(class_constructor) = &self.class_constructor {
            class_constructor.unparse(buf)?;
        }
        write!(buf, " ")?;
        if !self.implements.is_empty() {
            self.implements.unparse(buf)?;
            write!(buf, " ")?;
        }
        unparse_contained_declarations!(self.declarations, buf);
        unparse_ok!()
    }
}

/* Function declarations and components */

impl Unparse for FunctionDefinition {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if self.is_public {
            write!(buf, "pub ")?;
        }
        if let Some(modifier) = &self.modifier {
            modifier.unparse(buf)?;
            write!(buf, " ")?;
        }
        write!(buf, "fn ")?;
        if !self.generics.is_empty() {
            self.generics.unparse(buf)?;
            write!(buf, " ")?;
        }
        self.identifier.unparse(buf)?;
        self.parameters.unparse(buf)?;
        write!(buf, " ")?;
        self.return_type.unparse(buf)?;
        write!(buf, " ")?;
        self.body.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for OperatorFunctionDefinition {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if self.is_public {
            write!(buf, "pub ")?;
        }
        if let Some(modifier) = &self.modifier {
            modifier.unparse(buf)?;
            write!(buf, " ")?;
        }
        self.operator.unparse(buf)?;
        write!(buf, " ")?;
        self.parameters.unparse(buf)?;
        write!(buf, " ")?;
        self.return_type.unparse(buf)?;
        write!(buf, " ")?;
        self.body.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for PlatformFunctionDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if self.is_public {
            write!(buf, "pub ")?;
        }
        write!(buf, "platform fn ")?;
        if !self.generics.is_empty() {
            self.generics.unparse(buf)?;
            write!(buf, " ")?;
        }
        self.identifier.unparse(buf)?;
        self.parameters.unparse(buf)?;
        write!(buf, " ")?;
        self.return_type.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for InterfaceFunctionDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if let Some(modifier) = &self.modifier {
            modifier.unparse(buf)?;
            write!(buf, " ")?;
        }
        if !self.generics.is_empty() {
            self.parameters.unparse(buf)?;
            write!(buf, " ")?;
        }
        self.identifier.unparse(buf)?;
        self.parameters.unparse(buf)?;
        write!(buf, " ")?;
        self.return_type.unparse(buf)?;
        if let Some(body) = &self.body {
            write!(buf, " ")?;
            body.unparse(buf)?;
        }
        unparse_ok!()
    }
}

impl Unparse for InterfaceOperatorFunctionDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if let Some(modifier) = &self.modifier {
            modifier.unparse(buf)?;
            write!(buf, " ")?;
        }
        write!(buf, "op ")?;
        if !self.generics.is_empty() {
            self.generics.unparse(buf)?;
            write!(buf, " ")?;
        }
        self.operator.unparse(buf)?;
        write!(buf, " ")?;
        self.parameters.unparse(buf)?;
        write!(buf, " ")?;
        self.return_type.unparse(buf)?;
        if let Some(body) = &self.body {
            write!(buf, " ")?;
            body.unparse(buf)?;
        }
        unparse_ok!()
    }
}

/* Function Modifier */

impl Unparse for FunctionModifier {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use FunctionModifier::*;
        match self {
            Static => write!(buf, "static"),
            Cons => write!(buf, "cons"),
            Mut => write!(buf, "mut"),
            Ref => write!(buf, "ref"),
            MutRef => write!(buf, "mut ref"),
        }
    }
}

/* Function Body */

impl Unparse for FunctionBody {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use FunctionBody::*;
        match self {
            Equals(expression) => {
                write!(buf, "= ")?;
                expression.unparse(buf)
            }
            Block(body) => body.unparse(buf),
            Alias(identifier) => {
                write!(buf, "alias ")?;
                identifier.unparse(buf)
            }
        }
    }
}

/* Class components */

impl ListUnparse for ClassConstructor {
    fn prefix() -> &'static str {
        "("
    }

    fn separator() -> &'static str {
        ", "
    }

    fn suffix() -> &'static str {
        ")"
    }

    fn inner(&self) -> Vec<&dyn Unparse> {
        to_unparse_vec!(self.0)
    }
}

impl Unparse for ClassConstructorParameter {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use ClassConstructorParameter::*;
        match self {
            Property(property) => property.unparse(buf),
            Field(field) => field.unparse(buf),
        }
    }
}

impl Unparse for PropertyDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if self.is_mutable {
            write!(buf, "mut ")?;
        }
        self.identifier.unparse(buf)?;
        write!(buf, ": ")?;
        self.declared_type.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for FieldDeclaration {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        if self.is_mutable {
            write!(buf, "mut ")?;
        }
        write!(buf, "fld ")?;
        self.identifier.unparse(buf)?;
        write!(buf, ": ")?;
        self.declared_type.unparse(buf)?;
        unparse_ok!()
    }
}

/* Statements */

impl Unparse for BlockStatement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "{{\n")?;
        for statement in &self.statements {
            statement.unparse(buf)?;
        }
        if let Some(expression) = &self.expression {
            expression.unparse(buf)?;
            write!(buf, "\n")?;
        }
        write!(buf, "}}")?;
        unparse_ok!()
    }
}

impl Unparse for Statement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use Statement::*;
        match self {
            BlockStatement(statement) => statement.unparse(buf),
            CallStatement(call) => call.unparse(buf),
            VariableDeclarationStatement(declaration) => declaration.unparse(buf),
            AssignStatement(assign_statement) => assign_statement.unparse(buf),
            ReturnStatement(return_expression) => return_expression.unparse(buf),
            IfStatement(if_statement) => if_statement.unparse(buf),
            IfElseStatement(if_else_statement) => if_else_statement.unparse(buf),
            WhileStatement(while_statement) => while_statement.unparse(buf),
            ForStatement(for_statement) => for_statement.unparse(buf),
        }?;
        write!(buf, ";\n")?;
        unparse_ok!()
    }
}

impl Unparse for VariableDeclarationStatement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "let ")?;
        if self.is_mutable {
            write!(buf, "mut ")?;
        }
        self.identifier.unparse(buf)?;
        if let Some(declared_type) = &self.declared_type {
            write!(buf, ": ")?;
            declared_type.unparse(buf)?;
        }
        if let Some(initializer) = &self.initializer {
            write!(buf, " = ")?;
            initializer.unparse(buf)?;
        }
        unparse_ok!()
    }
}

impl Unparse for AssignStatement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        self.lhs.unparse(buf)?;
        write!(buf, " = ")?;
        self.rhs.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for CallStatement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        self.0.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for ReturnStatement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "return")?;
        if let Some(expression) = &self.0 {
            write!(buf, " ")?;
            expression.unparse(buf)?;
        }
        unparse_ok!()
    }
}

impl Unparse for IfStatement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "if ")?;
        self.condition.unparse(buf)?;
        write!(buf, " {{\n")?;
        self.then_block.unparse(buf)?;
        write!(buf, "\n}}")?;
        unparse_ok!()
    }
}

impl Unparse for IfElseStatement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "if ")?;
        self.condition.unparse(buf)?;
        write!(buf, " ")?;
        self.then_block.unparse(buf)?;
        write!(buf, " ")?;
        self.else_ifs.unparse(buf)?;
        write!(buf, "else ")?;
        self.else_block.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for ElseIfs {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        for if_statement in &self.0 {
            write!(buf, " else ")?;
            if_statement.unparse(buf)?;
            write!(buf, " ")?;
        }
        unparse_ok!()
    }
}

impl Unparse for WhileStatement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "while ")?;
        self.condition.unparse(buf)?;
        write!(buf, " ")?;
        self.body.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for ForStatement {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "for ")?;
        self.variable.unparse(buf)?;
        write!(buf, " in ")?;
        self.iterator.unparse(buf)?;
        write!(buf, " ")?;
        self.body.unparse(buf)?;
        unparse_ok!()
    }
}

/* Expressions */

impl Unparse for Expression {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use Expression::*;
        match self {
            Ternary(ternary) => ternary.unparse(buf),
            Binary(binary) => binary.unparse(buf),
            UnaryPrefix(prefix) => prefix.unparse(buf),
            UnarySuffix(suffix) => suffix.unparse(buf),
            Call(call) => call.unparse(buf),
            ObjectAccess(object_access) => object_access.unparse(buf),
            Literal(literal) => literal.unparse(buf),
            FullyQualifiedName(fully_qualified_name) => fully_qualified_name.unparse(buf),
            Closure(closure) => closure.unparse(buf),
        }
    }
}

impl Unparse for TernaryExpression {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        self.condition.unparse(buf)?;
        write!(buf, " ? ")?;
        self.true_expression.unparse(buf)?;
        write!(buf, " : ")?;
        self.false_expression.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for BinaryExpression {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        self.left.unparse(buf)?;
        write!(buf, " ")?;
        self.operator.unparse(buf)?;
        write!(buf, " ")?;
        self.right.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for PrefixExpression {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use PrefixUnaryOperator::*;
        match &self.operator {
            BorrowMut => write!(buf, "&mut "),
            Mut => write!(buf, "mut "),
            other => other.unparse(buf),
        }?;
        self.expression.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for SuffixExpression {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        self.expression.unparse(buf)?;
        self.operator.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for CallExpression {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        self.callee.unparse(buf)?;
        if let Some(turbo_fish) = &self.turbo_fish {
            turbo_fish.unparse(buf)?;
        }
        self.arguments.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for TurboFish {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "::")?;
        self.0.unparse(buf)?;
        unparse_ok!()
    }
}

impl ListUnparse for CallArguments {
    fn prefix() -> &'static str {
        "("
    }

    fn separator() -> &'static str {
        ", "
    }

    fn suffix() -> &'static str {
        ")"
    }

    fn inner(&self) -> Vec<&dyn Unparse> {
        to_unparse_vec!(self.0)
    }
}

impl Unparse for CallArgument {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        self.0.unparse(buf)
    }
}

impl Unparse for Closure {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        write!(buf, "{{ -> TODO }}")
    }
}

impl Unparse for ObjectAccess {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        self.receiver.unparse(buf)?;
        self.navigations.unparse(buf)?;
        unparse_ok!()
    }
}

impl Unparse for ObjectNavigations {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        for navigation in &self.0 {
            navigation.unparse(buf)?;
        }
        unparse_ok!()
    }
}

impl Unparse for ObjectNavigation {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use ObjectNavigation::*;
        match self {
            Index(expression) => {
                write!(buf, "[")?;
                expression.unparse(buf)?;
                write!(buf, "]")?;
            }
            Identifier(identifier) => {
                write!(buf, ".")?;
                identifier.unparse(buf)?;
            }
        }
        unparse_ok!()
    }
}

impl Unparse for Literal {
    fn unparse(&self, buf: &mut dyn std::fmt::Write) -> std::fmt::Result {
        use Literal::*;
        match self {
            Integer(i) => write!(buf, "{}", i),
            Long(l) => write!(buf, "{}", l),
            Double(d) => write!(buf, "{}", d),
            USize(u) => write!(buf, "{}", u),
            String(s) => write!(buf, "\"{}\"", s),
            Boolean(b) => write!(buf, "{}", b),
        }
    }
}
