use deimos::object_file::{DvmObjectFile, DvmPath};
use deimos::vm::constant::DvmConstant;
use deimos::vm::function::DvmFunction;
use deimos::vm::instruction::{Immediate, Instruction};
use deimos::vm::platform::get_std_lib_platform_functions;
use deimos::vm::source_code_location::SourceCodeLocation;
use deimos::vm::{dump_state, run_main_function, DvmContext, DvmState};
use std::rc::Rc;

fn main() {
    // Goal:
    // - write a single lib with a main()
    // - call the main fn
    //
    // ns greeter
    //
    // fn main() {
    //     println "Hello, World!"
    //     println foo()
    // }
    //
    // fn foo() = 42
    let mut greeter_object_file = DvmObjectFile::new(DvmPath::new("greeter", "greeter.dmo"));

    use Instruction::*;
    let main_instructions = vec![
        MoveImmediate {
            destination_register: 0,
            immediate: Immediate::Constant(DvmConstant::String(String::from("Hello, World!"))),
        },
        // prepare for call to println
        Push { source_register: 0 }, // save r0
        Push { source_register: 0 }, // arg0
        // call
        InvokeStaticPlatform {
            function_name: Rc::new(String::from("std::core::println")),
            source_code_location: SourceCodeLocation {
                source_file_name: String::from("greeter.dm"),
                line: 4,
                col: 5,
            },
        },
        // unwind from call
        Pop {
            destination_register: None,
        }, // arg0
        Pop {
            destination_register: Some(0), // restore r0
        },
        // prepare for call to foo
        Push { source_register: 0 }, // save r0
        InvokeStatic {
            function_name: Rc::new(String::from("greeter::foo")),
            source_code_location: SourceCodeLocation {
                source_file_name: String::from("greeter.dm"),
                line: 5,
                col: 13,
            },
        },
        Pop {
            destination_register: Some(0),
        }, // restore r0
        TakeReturnValue {
            destination_register: 1,
        },
        // call println again
        Push { source_register: 1 }, // save r1
        Push { source_register: 0 }, // save r0
        Push { source_register: 1 }, // arg0
        InvokeStaticPlatform {
            function_name: Rc::new(String::from("std::core::println")),
            source_code_location: SourceCodeLocation {
                source_file_name: String::from("greeter.dm"),
                line: 5,
                col: 5,
            },
        },
        Pop {
            destination_register: None,
        }, // arg0
        Pop {
            destination_register: Some(0), // restore r0
        },
        Pop {
            destination_register: Some(1), // restore r1
        },
        // return 0
        MoveImmediate {
            destination_register: 2,
            immediate: Immediate::Int(0),
        },
        SetReturnValue { source_register: 2 },
        Return, // explicit, not needed
    ];

    let main_function = DvmFunction::new(
        "greeter::main",
        &main_instructions,
        SourceCodeLocation {
            source_file_name: String::from("greeter.dm"),
            line: 3,
            col: 1,
        },
    );

    let foo_function_instructions = vec![
        MoveImmediate {
            destination_register: 0,
            immediate: Immediate::Int(42),
        },
        SetReturnValue { source_register: 0 },
        Return, // explicit, not needed
    ];

    let foo_function = DvmFunction::new(
        "greeter::foo",
        &foo_function_instructions,
        SourceCodeLocation {
            source_file_name: String::from("greeter.dm"),
            line: 8,
            col: 1,
        },
    );

    greeter_object_file.add_function(main_function);
    greeter_object_file.add_function(foo_function);

    let mut state = DvmState::new();
    let mut context = DvmContext::new("greeter::main");
    context.add_platform_functions(get_std_lib_platform_functions());
    greeter_object_file.load_to_context(&mut context, &|path| todo!());

    let exit_code = run_main_function(&mut state, &context);

    dump_state(&state, "After main!");

    if exit_code != 0 {
        std::process::exit(exit_code);
    }
}
